#include <iostream>
#include <math.h>
#include <complex>
#include <string>
#include <cstdlib>
#include <sys/param.h>
#include <sys/time.h>
#include <sys/times.h>
#include <sys/types.h>

//#include "mmeep_globals.h"
#include "mmeep_wrapper.h"
#include "mtpc_app.h"


using namespace std;


////////////////////////////////////////////////////////////////////////////////
//
// MMEEP_WRAPPER
//
////////////////////////////////////////////////////////////////////////////////


MMEEP_WRAPPER::MMEEP_WRAPPER(int id, double patch_length, double patch_width, 
    double recess, double depth): 
_id(id), 
_patch_length(patch_length),
_patch_width(patch_width), 
_recess(recess), 
_depth(depth) 
{

    // (1) initialize patch geometries; one time deal;
    initialize_geometrical_paramaters();

    // (2) some more debug info;
    problem_statement_paramaters();

    // (3) populate flux1r from the user provided refData.dat;
    read_user_refData();

    // (4) allocate memory for _frequency array used isnide findSpar();
    _frequency = new double[nfreq];
    _S11 = new double[nfreq];

    // (5) initializations; useful for debug too;
    _cost = -1.0;
    _cost_c1 = -1.0; 
    _cost_c2 = -1.0;
    _iFreqMin = -1;
    _center_freq = -1.0;
    _BW = -1.0;
}


void MMEEP_WRAPPER::initialize_geometrical_paramaters()
{
    // I put here what initially was in Jim's void inputGeo();
    // Note: for the time being I work with Jim's globals; not good programming
    // but, will move them into this class later;

    double dielHeight, stripWidth;
    double patchL, patchW;
    double recess, depth;


    printf("\nMMEEP WRAPPER instance #%d, initializing geometrical parameters...", _id);

    // (1) 2 hard-coded parameters;
    // param #1: hard-coded and fixed;
    // Here is declaration where one block is twice substrate thickness;
    dielHeight = 0.1575;
    blockCm = 2. * dielHeight; // #1

    // param #2: hard coded and fixed also;
    stripWidth = 0.451; 
    stripWidthBlk = stripWidth / blockCm; // #2


    // (2) four variables parameters;
    // param #3, param #4, param #5, param #6
    // initialized with the first value from their individual list of values,
    // which were populated wile reading from input filr;

    patchL = _patch_length; // e.g., 4.0; comes from user via Ctor;
    patchLBlk = patchL / blockCm; // #3 computed indirectly from user's patchL;

    patchW = _patch_width; // e.g., 4.9; comes from user via Ctor;
    patchWBlk = patchW / blockCm; // #4 computed indirectly from user's patchW;
  
    recess = _recess; // e.g., 1.0; comes from user via Ctor;
    rBlk = recess / blockCm; // #5 computed indirectly from user's recess;

    depth = _depth; // e.g., 0.2; comes from user via Ctor;
    dBlk = depth / blockCm; // #6 computed indirectly from user's depth;


    // (3) some more hard-coded values;
    portBlk = 2.;
    sourceBlk = 1.5;
    AddTlBlk = 1.;
    topSpaceBlk = 2.5; // tried 4.0, results identical
    XmaxBlk = 18.;
    YmaxBlk = 31.;
    PmlThickBlk = 1.;
}


void MMEEP_WRAPPER::problem_statement_paramaters()
{
    // I put here what initially was in Jim's void inputProbStat(void);
    // TODO: learn about the meaning of these; use variable names that are
    // more self-explanatory; move them as member variables of this class;

    double WLcm, freqMhz, dfMhz;

    freqMhz = 2400.; 

    WLcm = 30000. / freqMhz;
    WLblk = WLcm / blockCm;
    freqBlk = 1./WLblk;

    dfMhz = 600.; 
    dfBlk = dfMhz * blockCm / 30000.;

    nfreq = 101; 

    resolution = 10;
    stripThickBlk = 1.01 / resolution;

    epsR = 2.33; 
}

void MMEEP_WRAPPER::read_user_refData() 
{
    // Note: this used to be done inside findSpar(); I moved it here
    // because I need it for each instance of mmeep_wrapper;
    // TODO: just hardcode entire refData.dat file here to avoid opening
    // and closing files...

    // Note: make sure you free up this memory inside destructor of this class!
    // fyi, nfreq is 101, set inside problem_statement_paramaters(), called before 
    // this function;
    flux1r = new double[nfreq];


    // first, read in the flux1r[i] data
    int i = 0, j = 0;
    std::ifstream fin;
    fin.open("refData.dat");
    if (!fin) {
        std::cout<<"ERROR: input refData.dat file not found!";
        exit(1);
    }
    //std::cout<<"here is the data read in from the file\n";
    for (i=0; i<nfreq; i++) {
        fin>>j>>flux1r[i];
        // std::cout<<j<<"  "<<flux1r[i]<<"\n";
    }
    fin.close();
    // std::cout<<"End of data\n";
}

void MMEEP_WRAPPER::print_parameters_materials()
{
    // this is a clone of Jim's initial void printParameters(void)
    // inside materials.cpp, which I removed;

    // \cmt{ document parameters}
    std::cout<<stripThickBlk<<" stripThickBlk\n";
    std::cout<<topSpaceBlk<<" topSpaceBlk\n";
    std::cout<<PmlThickBlk<<" PmlThickBlk\n";
    std::cout<<blockCm<<" blockCm\n";
    std::cout<<WLblk<<" WLblk\n";
    std::cout<<freqBlk<<" freqBlk\n";
    std::cout<<dfBlk<<" dfBlk\n";
    std::cout<<resolution<<" resolution\n";
    std::cout<<zUpEps<<" zUpEps; "<<zUpAir<<" zUpAir\n";
    std::cout<<epsR<<" epsR\n";
}


void MMEEP_WRAPPER::do_meep_simulation_to_evaluate_solution()
{
    printf("\nEvaluating solution using one meep simulation instance...");

    // (1) run one meep simulation here;
    // basically, put together runMEEP() and findSpar();
    runMEEP();
    findSpar();
 
    // (2) calculate cost of this simulated solution; used for 
    // simulated annealing algo;
    calculate_cost_function_for_SA();
 
}


double MMEEP_WRAPPER::calculate_cost_function_for_SA()
{
    // the two objectives are:
    // 1) get center frequency as close as possible to f_c = 2400 MHz; that is
    // we want to minimize distance to f_c;
    // 2) maximize bandwidth, BW; assume an ideal bandwidth of 100 MHz; then,
    // we want to minimize "distance" between current bandwidth and that ideal one;
    // 
    // so, our overall cost, which we'll minimize, is:
    // cost = alpha * c1 + (1-alpha) * c2
    //      = alpha * abs(this_f_c - 2400) + (1-alpha) * abs(100 - BW)
    // and will use alpha = 0.5 to give equal weight to both cost components;
    // also, I should work with normalized c1,c2; to do that, I take ratio with respect to
    // and assumed fixed maximum (worst case) for each of c1 and c2;
    // I'll assume normalization factors of:
    // 300 for c1 (i.e., maximum that this_f_c can be away from ideal 2400 is 300)
    // 100 for c2 (i.e., maximum that this BW can be away from ideal 100 is 100)
    //
    // Note:
    // Center Frequency is stored in frequency[iFreqMin]
    // Bandwidth is stored in BW
    //
    // Examples of actual runs:
    // Center Frequency = 2406.000000 ; Bandwidth = 38.299463
    // Center Frequency = 2178.000000 ; Bandwidth = 56.503777
    // Center Frequency = 2178.000000 ; Bandwidth = 56.503777
    // Center Frequency = 2172.000000 ; Bandwidth = 59.863287
    
    double alpha = 0.5;
    _cost = 0.0;

    _cost_c1 = abs( 2400 - _center_freq);
    _cost_c2 = abs( 100 - _BW);
    _cost = alpha * _cost_c1/300 + (1 - alpha) * _cost_c2/100;
    //assert( _cost >= 0.0 && _cost <= 1.0);
    
    return ( _cost);
}
