#include <sys/time.h>
#include <string.h>

#include "mtpc_basics.h"


using namespace std;


////////////////////////////////////////////////////////////////////////////////
//
// DIRECTOR
//
////////////////////////////////////////////////////////////////////////////////

DIRECTOR::DIRECTOR() 
{
    _threads_count = 0;
    _worker_threads_count = 0;
    _cpus_count = 1;
    MTPC_MUTEX_INIT(&_mutex);
    char this_line[256];
    FILE *fp = fopen("/proc/cpuinfo", "r");
    if (fp) {
        _cpus_count = 0;
        while (fgets(this_line, 256, fp) != NULL) {
            if (strncmp(this_line,"processor", 9) == 0)
                _cpus_count ++;
        }
        if ( _cpus_count == 0) _cpus_count = 1;
        fclose(fp);
    }
}
DIRECTOR::~DIRECTOR()
{
}
bool DIRECTOR::incr_threads_count(bool this_is_worker)
{
    MTPC_MUTEX_LOCK(&_mutex);
    if (this_is_worker && _worker_threads_count >= _threads_count) {
        MTPC_MUTEX_UNLOCK(&_mutex);
        return false;
    }
    _threads_count++;
    if (this_is_worker) _worker_threads_count++;
    MTPC_MUTEX_UNLOCK(&_mutex);
    return true;
}
void DIRECTOR::decr_threads_count(bool this_is_worker)
{
    MTPC_MUTEX_LOCK(&_mutex);
    if (this_is_worker && _worker_threads_count > 0)
        _worker_threads_count--;
    if (_threads_count > 0) _threads_count--;
    MTPC_MUTEX_UNLOCK(&_mutex);
}

////////////////////////////////////////////////////////////////////////////////
//
// ONE_THREAD
//
////////////////////////////////////////////////////////////////////////////////

int ONE_THREAD::start_as_generator()
{
    if ( !g_director.incr_threads_count(false))
        return -1;
    int res = THREAD_CREATE( _thread, generator_thread_start, this);
    if ( res != 0) g_director.decr_threads_count(false);
    else _framework->add_generator( _thread);
    return res;
}
void *ONE_THREAD::generator_thread_start(void *arg) 
{
    // when a thread is created of type generator, we give it
    // this function which kicks in its execution;
    ONE_THREAD *ot = (ONE_THREAD*)arg;
    ot->get_framework()->run_generator_thread();
    g_director.decr_threads_count(false);
    return (void*) 0;
}

int ONE_THREAD::start_as_worker()
{
    if ( !g_director.incr_threads_count(true))
        return -1;
    int res = THREAD_CREATE( _thread, worker_thread_start, this);
    if ( res != 0) g_director.decr_threads_count(true);
    else _framework->add_worker( _thread);
    return res;
}
void *ONE_THREAD::worker_thread_start(void *arg) 
{
    // when a thread is created of type worker, we give it
    // this function which kicks in its execution;
    ONE_THREAD *ot = (ONE_THREAD*)arg;
    ot->get_framework()->run_worker_thread();
    g_director.decr_threads_count(true);
    return (void*) 0;
}
int ONE_THREAD::set_affinity_custom(int i)
{
    // create a cpu_set_t object representing a set of CPUs;
    // clear it and mark only CPU i as set;
    int res = 0;
    cpu_set_t cpuset;
    CPU_ZERO(&cpuset);
    CPU_SET(i, &cpuset);
    res = pthread_setaffinity_np( _thread, sizeof(cpu_set_t), &cpuset);
    return res;
}
