#include <iostream>
#include <stdlib.h>

#include <meep.hpp>

using namespace meep;

#include "mmeep_wrapper.h"


////////////////////////////////////////////////////////////////////////////////
//
// MMEEP_WRAPPER
//
////////////////////////////////////////////////////////////////////////////////


void MMEEP_WRAPPER::runMEEP(void)
{
    double xs,ysP,ysM,xp1,zP,zM;
    double freqMin,freqMax; 
    double finalTime;
  
    sx = 2.*PmlThickBlk + AddTlBlk + sourceBlk + portBlk + XmaxBlk;
    sy = 2.*PmlThickBlk + YmaxBlk;
    sz =    PmlThickBlk + 0.5 + topSpaceBlk;
    
    // \cmt{Check and report the geometry parameters}
    //master_printf("sx = %f , sy = %f , sz = %f\n",sx,sy,sz); 
    

    // \cmt{set up variables for eps function}
    zUpEps=0.5;
    zUpAir=zUpEps+stripThickBlk;
    

    // \cmt{print parameters for documentation/debugging}
    //printParameters();
    //host_mmeep_wrapper->print_parameters_materials(); // same thing as above;

    
    // \cmt{now, set up for meep run, including geometry report} 
    initialize mpi(); // mpi(argc,argv);
    
    // quiet=true;  to suppress all printing of information except error messages
    grid_volume v=vol3d(sx,sy,sz,resolution);
    

    // cris:
    // set up Patch Antenna in a different way than jim did in the original code;
    // that is because creating a new s structure passing the old "patch"
    // function was not working because patch would be a class member variable
    // not a simple global function; so, doing it here in the other way; structure
    // class has more constructors; see /home/cristinel/projects_antenna/meep/src/meep.hpp
    // for details;
    patch_material_function my_epsilon(this);
    structure *s;
    s = new structure(v, my_epsilon,
                      pml(PmlThickBlk,X)+pml(PmlThickBlk,Y)+pml(PmlThickBlk,Z,High),
                      meep::identity(), 0, 0.5);
    
 
    // cris:
    // set up loss for antenna dielectric in a different way than jim did in the original code;
    // this is for the same reason as above, in the case of "patch" function;
    loss_material_function my_conductivity(this);
    s->set_conductivity(Dx, my_conductivity);
    s->set_conductivity(Dy, my_conductivity);
    s->set_conductivity(Dz, my_conductivity);


    fields *f;
    f = new fields(s);


    // to see material geometry
    // f->output_hdf5(Dielectric,v.surroundings());
    // f0.output_hdf5(Dielectric,v.surroundings());
    // return;   

    // \cmt{Here, set up pointer for src which will be either cw or gauss}
    src_time *src = NULL;

    // \cmt{gaussian pulse src}
    if (dfBlk!=0) {
        freqMin=freqBlk-dfBlk/2.;
        freqMax=freqBlk+dfBlk/2.;

        finalTime=resolution/dfBlk; 
        //master_printf("%f is the final time using homemade relation\n",finalTime);

        src = new gaussian_src_time(freqBlk,dfBlk);
 
        finalTime=src->last_time();
        master_printf("%f is the last source time according to MEEP\n",finalTime);
    }
    
    //  \cmt{set the source location} 
    xs=PmlThickBlk+AddTlBlk;
    
    // \cmt{extent of source in y is 2/5 of strip width, centered underneath}
    ysP=(sy/2.)+stripWidthBlk/5.;
    ysM=(sy/2.)-stripWidthBlk/5.;
    volume src_volume(vec(xs,ysM,0),vec(xs,ysP,0.49));
    
    // \cmt{implement the source}
    f->add_volume_source(Ez,*src,src_volume);
    // f0.add_volume_source(Ez,*src,src_volume);

    // \cmt{define P1 and P2 flux planes (ports)}
    xp1=PmlThickBlk+AddTlBlk+sourceBlk;
  
    ysP=(sy/2.)+stripWidthBlk;
    ysM=(sy/2.)-stripWidthBlk;
  
    zP=1.0;
    zM=0.0;
  
    volume P1flux(vec(xp1,ysM,zM),vec(xp1,ysP,zP));
  
    dft_flux P1t = f->add_dft_flux_plane(P1flux,freqMin,freqMax,nfreq);
    // dft_flux P1r = f0.add_dft_flux_plane(P1flux,freqMin,freqMax,nfreq);


    // \cmt{adjust the final time and print}
    //master_printf("Final Time used is %f\n",finalTime);


    // (4)
    // \cmt{Run the test case}  ---------------------------------

    // debug;
    // cris: this is normally commented out; uncomment in order to make the
    // run short but without meaningful results;
    //finalTime = 0.2; // 0.2; 2300.; debug for 30.; 10.
    

    while (f->time() < finalTime) {
        f->step();
    }
    
    
    // \cmt{Save the test flux values}
    flux1t=P1t.flux();
    
    // End of test case ------------------------------------------


    // (5) clean-up; free allocated memory here; and return;
    delete s;
    delete f;
    if ( src != NULL) { 
        delete src; 
    }
    
    return;
}

////////////////////////////////////////////////////////////////////////////////
//
// 
//
////////////////////////////////////////////////////////////////////////////////

